from fruit.modules.step import Step

class FruitError(Exception):
    """Error class for aborting the target make"""
    pass

class Target(object):
    name: str = ""
    desc: str = ""
    __func : callable = None

    __steps: list = None

    def __init__(self, func: any, name: str, desc: str):
        """
        Initialize a new target from name description and callback
        
        Parameters
        ----------
        `func` : any
            Target function to call
        `name` : str
            Name of the target
        `desc` : str
            Description of the target
        """
        
        if type(name) is not str:
            raise TypeError("The target name must be a string!")

        if type(desc) is not str:
            raise TypeError("The target description must be a string!")
        
        if not callable(func):
            raise TypeError("The target function must be a callable!")

        self.name = name
        self.desc = desc
        self.__func = func

        self.__steps = []
    
    def add_step(self, step: Step):
        """
        Add a step, that is currently being executed to the target.
        
        Parameters
        ----------
        `step` : Step
            Step object
        """
        self.__steps.append(step)
    
    def count_steps(self) -> int:
        """
        Get the number of registered steps
        
        Returns
        -------
        int
            Number of steps
        """
        return len(self.__steps)
        

    def __call__(self):
        self.__func()


# Import console logging / printing functions
from fruit.modules.console import echo, error, warning

# Export decorators
from fruit.api.decorators import target, step

# Export exceptions
from fruit.api.api import abort

# Export the shell module
from fruit.api.shell import shell



import click

def echo(obj: any=None):
    """
    Write a string to the console.
    
    Parameters
    ----------
    `obj` : any
        Any object, that is compatible with the `print()` function.
    """
    if obj is not None:
        click.echo(obj)
    else:
        click.echo()

def warning(obj: any):
    click.secho(obj, fg='yellow')

def error(obj: any):
    click.secho(obj, fg='red')


"""
Module for implementing shell interactions within fruit
"""
import os
import fruit.modules.console as console
from fruit.globals import SHELLCHAR

def shell(cmd: str) -> int:
    """
    Execute the given shell string 
    
    Parameters
    ----------
    `cmd` : str
        Command string to execute
    
    Returns
    -------
    int
        Returncode of the shell
    """
    # Print the string to the console
    console.echo(SHELLCHAR + cmd)
    return os.system(cmd)



"""
This module contains python patterns to ease programming and usage.
"""
from __future__ import annotations
from typing import Optional

class Borg:
    """Singleton pattern Borg singleton pattern"""
    _shared_state = {}
    def __init__(self):
        self.__dict__ = self._shared_state

class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instance: Optional[Singleton] = None

    def __call__(self) -> Singleton:
        if self._instance is None:
            self._instance = super().__call__()
        return self._instance


"""
Fruit garden is the master collection singleton of the application
"""
from .patterns import SingletonMeta
from .target   import Target, FruitError
import fruit.modules.console as console

class Garden(metaclass=SingletonMeta):

    __targets : list = None
    __active_target : Target = None

    def __init__(self):

        # Initialize the collection the first time
        if self.__targets is None:
            self.__targets = []
    
    def add_target(self, target: Target):
        """
        Add a fruit target to the collection of targets
        
        Parameters
        ----------
        `target` : Target
            Target object
        """
        self.__targets.append(target)
    
    def make_target(self, target_name: str):
        """
        Execute the target with the selected name
        
        Parameters
        ----------
        `target_name` : str
            Target name to make
        """
        flt_target = list(filter(lambda trg: trg.name == target_name, self.__targets))
        
        if len(flt_target) < 1:
            raise ValueError("The target '{}' is not found!".format(target_name))
        else:
            cl, = flt_target
            try:
                self.__active_target = cl  # Set the currently active target
                cl()
            except FruitError as ferr:
                # Log the error message
                console.error("The make process was aborted. Reason:")
                console.error(str(ferr))
            finally:
                self.__active_target = None
    
    def active_target(self) -> Target:
        return self.__active_target


    def get_targets(self):
        for each_target in self.__targets:
            yield each_target


"""
Module for implementing the inner steps of a target
"""
import fruit.modules.console as console
from fruit.globals import FMT_STEPHEADER


class Step(object):
    name: str = ""  # Name of the step
    desc: str = ""  # Description of the step
    number: int = 0
    status: int = 1  # Error by default

    def __init__(self, name: str, desc: str, number: int):
        if type(name) is not str:
            raise TypeError("The step name must be a string!")
        if type(desc) is not str:
            raise TypeError("The step description bust be a string!")
        
        if type(number) is not int:
            raise TypeError("The step number must be an integer!")

        self.name = name
        self.desc = desc
        self.number = number
    
    def setstatus(self, value: int):
        """
        Set the status of the step to True (successful).
        """
        self.status = value
    
    def getstatus(self) -> int:
        return self.status
    
    def print_stephead(self):
        """
        Print a formatted step header to the console to separate steps from each other.
        """
        console.echo(FMT_STEPHEADER.format(number=self.number, name=self.name))

    



from fruit.modules.target import FruitError

def abort(reason: str):
    """
    Abort the target make from a step or a function
    
    Parameters
    ----------
    `reason` : str
        Abort reason
    
    Raises
    ------
    FruitError
        Error signal for target make 
    """
    raise FruitError(reason)


"""
Tratatata...
"""

import click
from tabulate import tabulate
from fruit.modules.fruitloader import load
from fruit.modules.garden import Garden
import fruit.modules.console as console

@click.group()
def cli():
    pass

@cli.command()
@click.argument('path', default='.')
def collect(path:str):
    """
    List the fruit targets in the given path.
    \b

    PATH (default: .) - Path of the directory or .py file to scan.
    """
    try:
        # Load the config file
        load(path)

        # List the targets
        tbl = []

        for trg in Garden().get_targets():
            tbl.append((trg.name, trg.desc))
        
        console.echo(tabulate(tbl, headers=['Target', 'Description']), )

    except Exception as err:
        console.error(str(err))



@cli.command()
@click.argument('target', required=True)
def make(target: str):
    """
    Make a fruit target from the parsed fruitconfig.py file.
    """
    try:
        console.echo()
        console.echo("🍓 Making target '{}'... ".format(target))
        console.echo()
        load('.')

        Garden().make_target(target)
    except Exception as err:
        console.error(str(err))



from fruit.modules.garden import Garden
from fruit.modules.target import Target
from fruit.modules.step   import Step

def target(func):
    garden = Garden()
    def wrapper():
        func()
    
    garden.add_target(Target(wrapper, func.__name__, func.__doc__))
    return wrapper

def step(func):
    def wrapper(*args, **kwargs):
        garden = Garden()
        next_nr = garden.active_target().count_steps() + 1
        new_step = Step(func.__name__, func.__doc__, next_nr)
        garden.active_target().add_step(new_step)
        # TODO: Maybe a context manager?
        new_step.print_stephead()
        func(*args, **kwargs)
    return wrapper



"""
This module implements the scan, loading and compiling of fruitconfig.py files.
"""

import os
import fruit.globals as glb

def obtain_config(path: str) -> str:
    """
    Get the full path an existing fruitconfig.py file
    
    Parameters
    ----------
    `path` : str
        Directory path to a containing directory or file path
    
    Returns
    -------
    str
        Full path of the config.py
    
    Raises
    ------
    FileNotFoundError
        The fruitconfig.py file is not found in the current directory
    FileNotFoundError
        The given path is invalid
    """
    if os.path.exists(path):
        if os.path.isdir(path):
            filepath = os.path.join(path, glb.FRUITCONFIG_NAME)
        else:
            filepath = path
        
        if os.path.exists(filepath) and os.path.isfile(filepath):
            return filepath
        else:
            raise FileNotFoundError("There is no fruitconfig.py in the current directory")
    else:
        raise FileNotFoundError("The given path is invalid!")

def compile_config(path: str):
    
    # Read the source code
    with open(path, 'r') as fp:
        source = fp.read()
    
    filename = os.path.basename(path)
    # Create a global namespace for the execution
    namespace = {}
    pyobj = compile(source=source, filename=filename, mode='exec')
    exec(pyobj, namespace, namespace)

def load(path: str):

    configpath = obtain_config(path)
    compile_config(configpath)